#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import sys
import pathlib
import logging

NAME = pathlib.Path(__file__).parent.name
VERSION = "2.1.0"
SCRAPER = f"{NAME} {VERSION}"

CHANNEL = "channel"
PLAYLIST = "playlist"
USER = "user"
ROOT_DIR = pathlib.Path(__file__).parent


class LessThanFilter(logging.Filter):
    def __init__(self, exclusive_maximum, name=""):
        super(LessThanFilter, self).__init__(name)
        self.max_level = exclusive_maximum

    def filter(self, record):
        """ non-zero return means we log this message """
        return 1 if record.levelno < self.max_level else 0


logger = logging.getLogger("youtube-scraper")
logger.setLevel(logging.NOTSET)
log_format = "[%(asctime)s] %(levelname)s:%(message)s"

stdout_handler = logging.StreamHandler(sys.stdout)
stdout_handler.setFormatter(logging.Formatter(log_format))
stdout_handler.addFilter(LessThanFilter(logging.WARNING))
stdout_handler.setLevel(logging.DEBUG)
logger.addHandler(stdout_handler)

stderr_handler = logging.StreamHandler(sys.stderr)
stdout_handler.setFormatter(logging.Formatter(log_format))
stderr_handler.setLevel(logging.WARNING)
logger.addHandler(stderr_handler)


class Youtube(object):
    def __init__(self):
        self.build_dir = None
        self.cache_dir = None
        self.api_key = None

    def update(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


YOUTUBE = Youtube()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

""" turn a regular build-folder into a low-quality one by re-encoding all videos.

    shortcut to call the FFMPEG youtube-dl hook we use in the scraper.
    videos are replaced.

    Usage: python youtube2zim/reencode_low_quality.py my/project/build/
"""

import sys
import json
import logging
import pathlib

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.DEBUG)
logger = logging.getLogger(__name__)


def main(build_path):
    build_dir = pathlib.Path(build_path)
    logger.info(f"about to re-encode video files in {build_dir}")

    if not build_dir.exists() or not build_dir.is_dir():
        logger.error(f"Build dir `{build_dir}` is not an existing folder.")
        sys.exit(1)

    # import youtube2zim hook
    sys.path = [str(pathlib.Path(__file__).parent.parent.resolve())] + sys.path
    from youtube2zim.converter import hook_youtube_dl_ffmpeg

    # retrieve source video_format
    with open(build_dir.joinpath("metadata.json"), "r") as fp:
        metadata = json.load(fp)
        video_format = metadata["video_format"]

    for video_id in build_dir.joinpath("videos").iterdir():
        if not video_id.is_dir():
            continue
        video_path = build_dir.joinpath("videos", video_id, f"video.{video_format}")
        logger.info(video_path)

        hook_youtube_dl_ffmpeg(
            video_format, True, {"status": "finished", "filename": str(video_path)}
        )

    logger.info("all done.")


if __name__ == "__main__":
    if len(sys.argv) != 2:
        logger.error("you must supply a path to a build folder")
        sys.exit(1)
    main(sys.argv[-1])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu


""" quick script to inject a function in ogv.js' -wasm.js scripts

    ogv.js dynamicaly loads .wasm scripts.
    it does it in two different manners, not always respecting its .base attribute
    .wasm scripts are binaries so they reside in /I/ inside ZIM files
    while javascript ones are in /-/.
    To easily circumvent this while keeping in working in HTML folder (no zim),
    we inject a `target=zim_fix_wasm_target(target)` in the -wasm.js files.
    We do it on the dist files as those files are not from the JS source of ogv.js
    but built with emscripten. Also, ogv.js toolchain is long to setup.

"""

import logging
import pathlib

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.DEBUG)
logger = logging.getLogger(__name__)


function_block = """
if (typeof zim_fix_wasm_target === 'undefined') {
    IS_IN_ZIM = self.location.href.indexOf("/-/") != -1 || self.location.href.indexOf("/I/") != -1 || self.location.href.indexOf("/A/") != -1;
    ZIM_IMG_NS = (IS_IN_ZIM) ? '../../../I/' : '';
    hasImageNamespacePrefix = function(target) { return target.indexOf("/I/") != -1; }
    hasMetaNamespacePrefix = function(target) { return target.indexOf("/-/") != -1; }
    changeNamespacePrefix = function(target, new_ns) { return target.replace("/-/", new_ns); }
    zim_fix_wasm_target = function(target) {
        console.debug("in-file zim_fix_wasm_target:", target);
        if (!IS_IN_ZIM) {
            console.debug("..not in zim");
            return target;
        }
        if (hasImageNamespacePrefix(target)) {
            // we already have a good path, leave it
        }
        else if (hasMetaNamespacePrefix(target)) {
            // we have a prefix, just replace it
            target = changeNamespacePrefix(target, "I");
        }
        else {
            // we lack the prefix, add it
            target = ZIM_IMG_NS + "assets/ogvjs/" + target;
        }
        console.debug("..target:", target);
        return target;
    }
  }

"""


def main():
    logger.info("about to add fix to ogv.js files to dynamicaly load wasm files in ZIM")

    root = pathlib.Path(__file__).parent
    ogvjs_path = root.joinpath("templates", "assets", "ogvjs")
    for fname in [
        "ogv-decoder-audio-opus-wasm.js",
        "ogv-decoder-audio-vorbis-wasm.js",
        # "ogv-decoder-video-av1-mt-wasm.js",
        "ogv-decoder-video-av1-wasm.js",
        "ogv-decoder-video-theora-wasm.js",
        # "ogv-decoder-video-vp8-mt-wasm.js",
        "ogv-decoder-video-vp8-wasm.js",
        # "ogv-decoder-video-vp9-mt-wasm.js",
        "ogv-decoder-video-vp9-wasm.js",
        "ogv-demuxer-ogg-wasm.js",
        "ogv-demuxer-webm-wasm.js",
    ]:
        fpath = ogvjs_path.joinpath(fname)

        with open(fpath, "r") as fp:
            content = fp.read()
        if "zim_fix_wasm_target" in content:
            logger.info(f"File `{fpath}` is already fixed!")
            continue
        else:
            logger.info(f"File `{fpath}` needs to be fixed.")

        # first pass, add function block
        vara_pos = content.index("var a;a")
        before = content[0:vara_pos]
        after = content[vara_pos:]
        content = before + function_block + after

        # second pass, add call to function
        locatefile_pos = content.index(".locateFile")
        bracket_pos = locatefile_pos + content[locatefile_pos:].index("}")
        before = content[0:bracket_pos]
        after = content[bracket_pos:]
        equal_pos = before[::-1].index("=")
        variable = before[::-1][equal_pos + 1 : equal_pos + 2]
        our_fix = ";{var}=zim_fix_wasm_target({var});".format(var=variable)

        # nfpath = fpath.parent.joinpath(fpath.name + ".tmp")
        with open(fpath, "w") as fp:
            fp.write(before + our_fix + after)

    logger.info("fixing videosjs-ogvjs.js")
    plugin_path = root.joinpath("templates", "assets", "videojs-ogvjs.js")
    with open(plugin_path, "r") as fp:
        content = fp.read()

    content = content.replace(
        "_OGVLoader2['default'].base = options.base;",
        "_OGVLoader2['default'].base = ZIM_META_NS + options.base;",
    )
    content = content.replace(
        "return type.indexOf('/ogg') !== -1 ? 'maybe' : '';",
        "return (type.indexOf('/webm') !== -1 || type.indexOf('/ogg') !== -1) ? 'maybe' : '';",
    )

    with open(plugin_path, "w") as fp:
        fp.write(content)

    logger.info("hack video.min.js (TEMP FIX to work aroung Qt bug in reader)")
    videojs_path = root.joinpath("templates", "assets", "videojs", "video.min.js")
    with open(videojs_path, "r") as fp:
        content = fp.read()

    content = content.replace(";return 0!==e?(t=", ";return (0!==e || IS_IN_ZIM)?(t=")

    with open(videojs_path, "w") as fp:
        fp.write(content)

    logger.info("all done.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import re
import json
import locale
import gettext
import colorsys

import PIL
import iso639
import requests
import colorthief
from slugify import slugify
from resizeimage import resizeimage

from .constants import ROOT_DIR


def get_slug(text, js_safe=True):
    """ slug from text to build URL parts """
    if js_safe:
        return slugify(text, regex_pattern=r"[^-a-z0-9_]+").replace("-", "_")
    return slugify(text)


def clean_text(text):
    """ cleaned-down version of text as Youtube is very permissive with descriptions """
    return text.strip().replace("\n", " ").replace("\r", " ")


def save_json(cache_dir, key, data):
    """ save JSON collection to path """
    with open(cache_dir.joinpath(f"{key}.json"), "w") as fp:
        json.dump(data, fp, indent=4)


def load_json(cache_dir, key):
    """ load JSON collection from path or None """
    fname = cache_dir.joinpath(f"{key}.json")
    if not fname.exists():
        return None
    try:
        with open(fname, "r") as fp:
            return json.load(fp)
    except Exception:
        return None


def save_file(url, fpath):
    """ download a binary file from its URL """
    req = requests.get(url)
    req.raise_for_status()
    if not fpath.parent.exists():
        fpath.parent.mkdir(exist_ok=True)
    with open(fpath, "wb") as fp:
        fp.write(req.content)


def get_colors(image_path, use_palette=True):
    """ (main, secondary) HTML color codes from an image path """

    def rgb_to_hex(r, g, b):
        """ hexadecimal HTML-friendly color code for RGB tuple """
        return "#{}{}{}".format(*[str(hex(x)[2:]).zfill(2) for x in (r, g, b)]).upper()

    def solarize(r, g, b):
        # calculate solarized color for main
        h, l, s = colorsys.rgb_to_hls(float(r) / 256, float(g) / 256, float(b) / 256)
        r2, g2, b2 = [int(x * 256) for x in colorsys.hls_to_rgb(h, 0.95, s)]
        return r2, g2, b2

    ct = colorthief.ColorThief(image_path)

    if use_palette:
        # extract two main colors from palette, solarizing second as background
        palette = ct.get_palette(color_count=2, quality=1)

        # using the first two colors of the palette?
        mr, mg, mb = palette[0]
        sr, sg, sb = solarize(*palette[1])
    else:
        # extract main color from image and solarize it as background
        mr, mg, mb = ct.get_color(quality=1)
        sr, sg, sb = solarize(mr, mg, mb)

    return rgb_to_hex(mr, mg, mb), rgb_to_hex(sr, sg, sb)


def resize_image(fpath, width, height=None, to=None, method="width"):
    """ resize an image file (dimensions)

        methods: width, height, cover """
    with open(str(fpath), "rb") as fp:
        with PIL.Image.open(fp) as image:
            if method == "width":
                resized = resizeimage.resize(method, image, width)
            elif method == "height":
                resized = resizeimage.resize(method, image, height=height)
            else:
                resized = resizeimage.resize(method, image, [width, height])
    kwargs = {"JPEG": {"quality": 100}}
    resized.save(
        str(to) if to is not None else fpath, image.format, **kwargs.get(image.format)
    )


def is_hex_color(text):
    """ whether supplied text is a valid hex-formated color code """
    return re.search(r"^#(?:[0-9a-fA-F]{3}){1,2}$", text)


def nicer_args_join(args):
    """ slightly better concateated list of subprocess args for display """
    nargs = args[0:1]
    for arg in args[1:]:
        nargs.append(arg if arg.startswith("-") else '"{}"'.format(arg))
    return " ".join(nargs)


def get_language_details(iso_639_3):
    """ dict container iso639-2, name and native name for an iso-639-3 code """
    non_iso_langs = {
        "zh-Hans": {
            "code": "zh-Hans",
            "iso-639-1": "zh",
            "english": "Simplified Chinese",
            "native": "简化字",
        },
        "zh-Hant": {
            "code": "zh-Hant",
            "iso-639-1": "zh",
            "english": "Traditional Chinese",
            "native": "正體字",
        },
        "iw": {"code": "iw", "iso-639-1": "he", "english": "Hebrew", "native": "עברית"},
        "es-419": {
            "code": "es-419",
            "iso-639-1": "es-419",
            "english": "Spanish",
            "native": "Español",
        },
        "multi": {
            "code": "mul",
            "iso-639-1": "en",
            "english": "Multiple Languages",
            "native": "Multiple Languages",
        },
    }

    try:
        return (
            non_iso_langs.get(iso_639_3)
            if iso_639_3 in non_iso_langs.keys()
            else {
                "code": iso_639_3,
                "iso-639-1": iso639.to_iso639_1(iso_639_3),
                "english": iso639.to_name(iso_639_3),
                "native": iso639.to_native(iso_639_3),
            }
        )
    except iso639.NonExistentLanguageError:
        return {
            "code": iso_639_3,
            "iso_639_3": iso_639_3,
            "english": iso_639_3,
            "native": iso_639_3,
        }


def setlocale(locale_name):
    computed = locale.setlocale(locale.LC_ALL, (locale_name.split(".")[0], "UTF-8"))
    gettext.bindtextdomain("messages", str(ROOT_DIR.joinpath("locale")))
    gettext.textdomain("messages")
    return computed



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

"""
    create project on Google Developer console
    Add Youtube Data API v3 to it
    Create credentials (Other non-UI, Public Data)
"""

import os
import json
import locale
import shutil
import datetime
from pathlib import Path
import concurrent.futures
from functools import partial
from gettext import gettext as _

import jinja2
import youtube_dl
from dateutil import parser as dt_parser
from babel.dates import format_date

from .zim import ZimInfo, make_zim_file
from .utils import (
    clean_text,
    resize_image,
    load_json,
    save_json,
    get_slug,
    save_file,
    get_colors,
    is_hex_color,
    get_language_details,
    setlocale,
)
from .youtube import (
    get_channel_json,
    credentials_ok,
    Playlist,
    get_channel_playlists_json,
    get_videos_json,
    get_videos_authors_info,
    save_channel_branding,
    skip_deleted_videos,
)
from .converter import hook_youtube_dl_ffmpeg
from .constants import logger, ROOT_DIR, CHANNEL, PLAYLIST, USER


class Youtube2Zim(object):
    def __init__(
        self,
        collection_type,
        youtube_id,
        api_key,
        video_format,
        low_quality,
        nb_videos_per_page,
        all_subtitles,
        autoplay,
        output_dir,
        no_zim,
        fname,
        debug,
        keep_build_dir,
        skip_download,
        max_concurrency,
        youtube_store,
        language,
        locale_name,
        tags,
        title=None,
        description=None,
        creator=None,
        publisher=None,
        name=None,
        profile_image=None,
        banner_image=None,
        main_color=None,
        secondary_color=None,
        only_test_branding=None,
    ):
        # data-retrieval info
        self.collection_type = collection_type
        self.youtube_id = youtube_id
        self.api_key = api_key

        # video-encoding info
        self.video_format = video_format
        self.low_quality = low_quality

        # options & zim params
        self.nb_videos_per_page = nb_videos_per_page
        self.all_subtitles = all_subtitles
        self.autoplay = autoplay
        self.fname = fname
        self.language = language
        self.tags = [t.strip() for t in tags.split(",")]
        self.title = title
        self.description = description
        self.creator = creator
        self.publisher = publisher
        self.name = name
        self.profile_image = profile_image
        self.banner_image = banner_image
        self.main_color = main_color
        self.secondary_color = secondary_color

        # process-related
        self.output_dir = Path(output_dir).expanduser().resolve()
        self.playlists = []
        self.uploads_playlist_id = None
        self.videos_ids = []
        self.main_channel_id = None  # use for branding
        self.only_test_branding = only_test_branding

        # debug/devel options
        self.no_zim = no_zim
        self.debug = debug
        self.keep_build_dir = keep_build_dir
        self.skip_download = skip_download
        self.max_concurrency = max_concurrency

        self.build_dir = self.output_dir.joinpath("build")

        # store ZIM-related info
        self.zim_info = ZimInfo(
            language=language,
            tags=tags,
            title=title,
            description=description,
            creator=creator,
            publisher=publisher,
            name=name,
        )

        # update youtube credentials store
        youtube_store.update(
            build_dir=self.build_dir, api_key=self.api_key, cache_dir=self.cache_dir
        )

        # set and record locale for translations
        locale_name = locale_name or get_language_details(self.language)["iso-639-1"]
        try:
            self.locale = setlocale(locale_name)
        except locale.Error:
            logger.error(
                f"No locale for {locale_name}. Use --locale to specify it. defaulting to en_US"
            )
            self.locale = setlocale("en")

    @property
    def root_dir(self):
        return ROOT_DIR

    @property
    def templates_dir(self):
        return self.root_dir.joinpath("templates")

    @property
    def assets_src_dir(self):
        return self.templates_dir.joinpath("assets")

    @property
    def assets_dir(self):
        return self.build_dir.joinpath("assets")

    @property
    def channels_dir(self):
        return self.build_dir.joinpath("channels")

    @property
    def cache_dir(self):
        return self.build_dir.joinpath("cache")

    @property
    def videos_dir(self):
        return self.build_dir.joinpath("videos")

    @property
    def profile_path(self):
        return self.build_dir.joinpath("profile.jpg")

    @property
    def banner_path(self):
        return self.build_dir.joinpath("banner.jpg")

    @property
    def is_user(self):
        return self.collection_type == USER

    @property
    def is_channel(self):
        return self.collection_type == CHANNEL

    @property
    def is_playlist(self):
        return self.collection_type == PLAYLIST

    @property
    def is_single_channel(self):
        if self.is_channel or self.is_user:
            return True
        return len(list(set([pl.creator_id for pl in self.playlists]))) == 1

    @property
    def sorted_playlists(self):
        """ sorted list of playlists (by title) but with Uploads one at first if any """
        if len(self.playlists) < 2:
            return self.playlists

        sorted_playlists = sorted(self.playlists, key=lambda x: x.title)
        index = 0
        # make sure our Uploads, special playlist is first
        if self.uploads_playlist_id:
            try:
                index = [
                    index
                    for index, p in enumerate(sorted_playlists)
                    if p.playlist_id == self.uploads_playlist_id
                ][-1]
            except Exception:
                index = 0
        return (
            [sorted_playlists[index]]
            + sorted_playlists[0:index]
            + sorted_playlists[index + 1 :]
        )

    def run(self):
        """ execute the scrapper step by step """
        logger.info(
            f"starting youtube scraper for {self.collection_type}#{self.youtube_id}"
        )

        logger.info("preparing build folder at {}".format(self.build_dir.resolve()))
        if not self.keep_build_dir and self.build_dir.exists():
            shutil.rmtree(self.cache_dir, ignore_errors=True)
            shutil.rmtree(self.build_dir)
        self.make_build_folder()

        logger.info("testing Youtbe credentials")
        if not credentials_ok():
            raise ValueError("Unable to connect to Youtube API v3. check `API_KEY`.")

        # fail early if supplied branding files are missing
        self.check_branding_values()

        logger.info("compute playlists list to retrieve")
        self.extract_playlists()

        logger.info(
            ".. {} playlists:\n   {}".format(
                len(self.playlists),
                "\n   ".join([p.playlist_id for p in self.playlists]),
            )
        )

        if self.only_test_branding:
            logger.info("update general metadata")
            self.update_metadata()

            logger.info("building fake homepage with branding")
            self.make_test_html_file()

            logger.info("done.")
            return

        logger.info("compute list of videos")
        self.extract_videos_list()
        logger.info(".. {} videos.".format(len(self.videos_ids)))

        # download videos (and recompress)
        logger.info(
            f"downloading all videos, subtitles and thumbnails (concurrency={self.max_concurrency})"
        )
        logger.info(f"  format: {self.video_format}")
        logger.info(f"  recompress: {self.low_quality}")
        logger.info(f"  generated-subtitles: {self.all_subtitles}")
        if not self.skip_download:
            succeeded, failed = self.download_video_files(
                max_concurrency=self.max_concurrency
            )
            if failed:
                logger.error(f"{len(failed)} video(s) failed to download: {failed}")
                if len(failed) >= len(succeeded):
                    logger.critical("More than half of videos failed. exiting")
                    raise IOError("Too much videos failed to download")

        logger.info("retrieve channel-info for all videos (author details)")
        get_videos_authors_info(succeeded)

        logger.info("download all author's profile pictures")
        self.download_authors_branding()

        logger.info("update general metadata")
        self.update_metadata()

        logger.info("creating HTML files")
        self.make_html_files(succeeded)

        # make zim file
        if not self.no_zim:
            period = datetime.datetime.now().strftime("%Y-%m")
            self.fname = Path(self.fname if self.fname else f"{self.name}_{period}.zim")
            logger.info("building ZIM file")
            print(self.zim_info.to_zimwriterfs_args())
            make_zim_file(self.build_dir, self.output_dir, self.fname, self.zim_info)
            logger.info("removing HTML folder")
            if not self.keep_build_dir:
                shutil.rmtree(self.build_dir, ignore_errors=True)

        logger.info("all done!")

    def make_build_folder(self):
        """ prepare build folder before we start downloading data """

        # create build folder
        os.makedirs(self.build_dir, exist_ok=True)

        # copy assets
        if self.assets_dir.exists():
            shutil.rmtree(self.assets_dir)
        shutil.copytree(self.assets_src_dir, self.assets_dir)

        # cache folder to store youtube-api results
        self.cache_dir.mkdir(exist_ok=True)

        # make videos placeholder
        self.videos_dir.mkdir(exist_ok=True)

        # make channels placeholder (profile files)
        self.channels_dir.mkdir(exist_ok=True)

    def check_branding_values(self):
        """ checks that user-supplied images and colors are valid (so to fail early)

            Images are checked for existence or downloaded then resized
            Colors are check for validity """

        # skip this step if none of related values were supplied
        if not sum(
            [
                bool(x)
                for x in (
                    self.profile_image,
                    self.banner_image,
                    self.main_color,
                    self.secondary_color,
                )
            ]
        ):
            return
        logger.info("checking your branding files and values")
        if self.profile_image:
            if self.profile_image.startswith("http"):
                save_file(self.profile_image, self.profile_path)
            else:
                if not self.profile_image.exists():
                    raise IOError(
                        f"--profile image could not be found: {self.profile_image}"
                    )
                shutil.move(self.profile_image, self.profile_path)
            resize_image(self.profile_path, width=100, height=100, method="thumbnail")
        if self.banner_image:
            if self.banner_image.startswith("http"):
                save_file(self.banner_image, self.banner_path)
            else:
                if not self.banner_image.exists():
                    raise IOError(
                        f"--banner image could not be found: {self.banner_image}"
                    )
                shutil.move(self.banner_image, self.banner_path)
            resize_image(self.banner_path, width=1060, height=175, method="thumbnail")

        if self.main_color and not is_hex_color(self.main_color):
            raise ValueError(
                f"--main-color is not a valid hex color: {self.main_color}"
            )

        if self.secondary_color and not is_hex_color(self.secondary_color):
            raise ValueError(
                f"--secondary_color-color is not a valid hex color: {self.secondary_color}"
            )

    def extract_playlists(self):
        """ prepare a list of Playlist from user request

            USER: we fetch the hidden channel associate to it
            CHANNEL (and USER): we grab all playlists + `uploads` playlist
            PLAYLIST: we retrieve from the playlist Id(s) """

        if self.is_user or self.is_channel:
            if self.is_user:
                # youtube_id is a Username, fetch actual channelId through channel
                channel_json = get_channel_json(self.youtube_id, for_username=True)
            else:
                # youtube_id is a channelId
                channel_json = get_channel_json(self.youtube_id)

            self.main_channel_id = channel_json["id"]

            # retrieve list of playlists for that channel
            playlist_ids = [
                p["id"] for p in get_channel_playlists_json(self.main_channel_id)
            ]
            # we always include uploads playlist (contains everything)
            playlist_ids += [
                channel_json["contentDetails"]["relatedPlaylists"]["uploads"]
            ]
            self.uploads_playlist_id = playlist_ids[-1]
        elif self.is_playlist:
            playlist_ids = self.youtube_id.split(",")
            self.main_channel_id = Playlist.from_id(playlist_ids[0]).creator_id
        else:
            raise NotImplementedError("unsupported collection_type")

        self.playlists = [
            Playlist.from_id(playlist_id) for playlist_id in list(set(playlist_ids))
        ]

    def extract_videos_list(self):
        all_videos = load_json(self.cache_dir, "videos")
        if all_videos is None:
            all_videos = {}

            # we only return video_ids that we'll use later on. per-playlist JSON stored
            for playlist in self.playlists:
                all_videos.update(
                    {
                        v["contentDetails"]["videoId"]: v
                        for v in filter(
                            skip_deleted_videos, get_videos_json(playlist.playlist_id)
                        )
                    }
                )

            # self.videos_ids = list(set(videos_ids))
            save_json(self.cache_dir, "videos", all_videos)
        self.videos_ids = [*all_videos.keys()]  # unpacking so it's subscriptable

    def download_video_files(self, max_concurrency):

        # prepare options which are shared with every downloader
        options = {
            "cachedir": self.videos_dir,
            "writethumbnail": True,
            "write_all_thumbnails": True,
            "writesubtitles": True,
            "allsubtitles": True,
            "subtitlesformat": "vtt",
            "keepvideo": False,
            "ignoreerrors": False,
            "retries": 20,
            "fragment-retries": 50,
            "skip-unavailable-fragments": True,
            # "external_downloader": "aria2c",
            # "external_downloader_args": ["--max-tries=20", "--retry-wait=30"],
            "outtmpl": str(self.videos_dir.joinpath("%(id)s", "video.%(ext)s")),
            "preferredcodec": self.video_format,
            "format": "{fmt}/best".format(fmt=self.video_format),
            "progress_hooks": [
                partial(hook_youtube_dl_ffmpeg, self.video_format, self.low_quality)
            ],
        }
        if self.all_subtitles:
            options.update({"writeautomaticsub": True})

        # find number of actuall parallel workers
        nb_videos = len(self.videos_ids)
        concurrency = nb_videos if nb_videos < max_concurrency else max_concurrency

        # short-circuit concurency if we have only one thread (can help debug)
        if concurrency <= 1:
            return self.download_video_files_batch(options, self.videos_ids)

        # prepare out videos_ids batches
        def get_slot():
            n = 0
            while True:
                yield n
                n += 1
                if n >= concurrency:
                    n = 0

        batches = [[] for _ in range(0, concurrency)]
        slot = get_slot()
        for video_id in self.videos_ids:
            batches[next(slot)].append(video_id)

        overall_succeeded = []
        overall_failed = []
        # execute the batches concurrently
        with concurrent.futures.ThreadPoolExecutor(max_workers=concurrency) as executor:
            fs = [
                executor.submit(self.download_video_files_batch, options, videos_ids)
                for videos_ids in batches
            ]
            done, not_done = concurrent.futures.wait(
                fs, return_when=concurrent.futures.ALL_COMPLETED
            )

            # we have some `not_done` batches, indicating errors within
            if not_done:
                logger.critical(
                    "Not all video-processing batches completed. Cancelling…"
                )
                for future in not_done:
                    exc = future.exception()
                    if exc:
                        logger.exception(exc)
                        raise exc

            # retrieve our list of successful/failed video_ids
            for future in done:
                succeeded, failed = future.result()
                overall_succeeded += succeeded
                overall_failed += failed

        # remove left-over files for failed downloads
        logger.debug(f"removing left-over files of {len(overall_failed)} failed videos")
        for video_id in overall_failed:
            shutil.rmtree(self.videos_dir.joinpath(video_id), ignore_errors=True)

        return overall_succeeded, overall_failed

    @staticmethod
    def download_video_files_batch(options, videos_ids):

        succeeded = []
        failed = []
        with youtube_dl.YoutubeDL(options) as ydl:
            for video_id in videos_ids:
                try:
                    ydl.download([video_id])
                    succeeded.append(video_id)
                except youtube_dl.utils.DownloadError:
                    failed.append(video_id)
        return succeeded, failed

    def download_authors_branding(self):
        videos_channels_json = load_json(self.cache_dir, "videos_channels")
        uniq_channel_ids = list(
            set([chan["channelId"] for chan in videos_channels_json.values()])
        )
        for channel_id in uniq_channel_ids:
            save_channel_branding(self.channels_dir, channel_id, save_banner=False)

    def update_metadata(self):
        # we use title, description, profile and banner of channel/user
        # or channel of first playlist
        main_channel_json = get_channel_json(self.main_channel_id)
        save_channel_branding(self.channels_dir, self.main_channel_id, save_banner=True)

        # if a single playlist was requested, use if for names;
        # otherwise, use main_channel's details.
        auto_title = (
            self.playlists[0].title
            if self.is_playlist and len(self.playlists) == 1
            else main_channel_json["snippet"]["title"].strip()
        )
        auto_description = (
            clean_text(self.playlists[0].description)
            if self.is_playlist and len(self.playlists) == 1
            else clean_text(main_channel_json["snippet"]["description"])
        )
        self.title = self.title or auto_title or "-"
        self.description = self.description or auto_description or "-"

        if self.creator is None:
            if self.is_single_channel:
                self.creator = _("Youtube Channel “{title}”").format(
                    title=main_channel_json["snippet"]["title"]
                )
            else:
                self.creator = _("Youtube Channels")
        self.publisher = self.publisher or "Kiwix"

        self.tags = self.tags or ["youtube"]
        if "_videos:yes" not in self.tags:
            self.tags.append("_videos:yes")

        self.zim_info.update(
            title=self.title,
            description=self.description,
            creator=self.creator,
            publisher=self.publisher,
            name=self.name,
            tags=self.tags,
        )

        # copy our main_channel branding into /(profile|banner).jpg if not supplied
        if not self.profile_path.exists():
            shutil.copy(
                self.channels_dir.joinpath(self.main_channel_id, "profile.jpg"),
                self.profile_path,
            )
        if not self.banner_path.exists():
            shutil.copy(
                self.channels_dir.joinpath(self.main_channel_id, "banner.jpg"),
                self.banner_path,
            )

        # set colors from images if not supplied
        if self.main_color is None or self.secondary_color is None:
            profile_main, profile_secondary = get_colors(self.profile_path)
        self.main_color = self.main_color or profile_main
        self.secondary_color = self.secondary_color or profile_secondary

        resize_image(
            self.profile_path,
            width=48,
            height=48,
            method="thumbnail",
            to=self.build_dir.joinpath("favicon.jpg"),
        )

    def make_test_html_file(self):

        video_id = "aaaabbbccddd"
        video_dir = self.videos_dir.joinpath(video_id)
        video_dir.mkdir(exist_ok=True)
        shutil.copy(
            self.assets_dir.joinpath("sample.jpg"), video_dir.joinpath("video.jpg")
        )

        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(str(self.templates_dir)), autoescape=True
        )

        html = env.get_template("home.html").render(
            playlists=self.sorted_playlists,
            video_format=self.video_format,
            title=self.title,
            description=self.description,
            color=self.main_color,
            background_color=self.secondary_color,
            page_label=_("Page {current}/{total}"),
            back_label=_("Back to top"),
        )
        with open(self.build_dir.joinpath("home.html"), "w", encoding="utf-8") as fp:
            fp.write(html)

        with open(self.assets_dir.joinpath("db.js"), "w", encoding="utf-8") as fp:
            fp.write(
                env.get_template("assets/db.js").render(
                    NB_VIDEOS_PER_PAGE=self.nb_videos_per_page
                )
            )

        with open(self.assets_dir.joinpath("data.js"), "w", encoding="utf-8") as fp:
            videos = [
                {
                    "id": video_id,
                    "title": "A Sample Title",
                    "slug": "a-sample-title",
                    "description": "A sample description of video",
                    "subtitles": [],
                    "thumbnail": str(Path("videos").joinpath(video_id, "video.jpg")),
                }
                for _ in range(0, 6)
            ]
            fp.write(
                "var json_{slug} = {json_str};\n".format(
                    slug=self.sorted_playlists[0].slug,
                    json_str=json.dumps(videos, indent=4),
                )
            )

    def make_html_files(self, actual_videos_ids):
        """ make up HTML structure to read the content

        /home.html                                  Homepage

        for each video:
            - <slug-title>.html                     HTML article
            - videos/<videoId>/video.<ext>          video file
            - videos/<videoId>/video.<lang>.vtt     subtititle(s)
            - videos/<videoId>/video.jpg            template
        """

        def is_present(video):
            """ whether this video has actually been succeffuly downloaded """
            return video["contentDetails"]["videoId"] in actual_videos_ids

        def get_subtitles(video_id):
            video_dir = self.videos_dir.joinpath(video_id)
            languages = [
                x.stem.split(".")[1]
                for x in video_dir.iterdir()
                if x.is_file() and x.name.endswith(".vtt")
            ]

            return [get_language_details(language) for language in languages]

        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(str(self.templates_dir)), autoescape=True
        )

        videos = load_json(self.cache_dir, "videos").values()
        # filter videos so we only include the ones we could retrieve
        videos = list(filter(is_present, videos))
        videos_channels = load_json(self.cache_dir, "videos_channels")
        for video in videos:
            video_id = video["contentDetails"]["videoId"]
            title = video["snippet"]["title"]
            slug = get_slug(title)
            description = video["snippet"]["description"]
            publication_date = dt_parser.parse(
                video["contentDetails"]["videoPublishedAt"]
            )
            author = videos_channels[video_id]
            subtitles = get_subtitles(video_id)
            video_url = f"https://www.youtube.com/watch?v={video_id}"

            html = env.get_template("article.html").render(
                video_id=video_id,
                video_format=self.video_format,
                author=author,
                title=title,
                description=description,
                date=format_date(publication_date, format="medium", locale=self.locale),
                subtitles=subtitles,
                url=video_url,
                channel_id=video["snippet"]["channelId"],
                color=self.main_color,
                background_color=self.secondary_color,
                autoplay=self.autoplay,
            )
            with open(
                self.build_dir.joinpath(f"{slug}.html"), "w", encoding="utf-8"
            ) as fp:
                fp.write(html)

        # build homepage
        html = env.get_template("home.html").render(
            playlists=self.playlists,
            video_format=self.video_format,
            title=self.title,
            description=self.description,
            color=self.main_color,
            background_color=self.secondary_color,
            page_label=_("Page {current}/{total}"),
            back_label=_("Back to top"),
        )
        with open(self.build_dir.joinpath("home.html"), "w", encoding="utf-8") as fp:
            fp.write(html)

        # rewrite app.js including `format`
        with open(self.assets_dir.joinpath("app.js"), "w", encoding="utf-8") as fp:
            fp.write(
                env.get_template("assets/app.js").render(video_format=self.video_format)
            )

        # rewrite app.js including `pagination`
        with open(self.assets_dir.joinpath("db.js"), "w", encoding="utf-8") as fp:
            fp.write(
                env.get_template("assets/db.js").render(
                    NB_VIDEOS_PER_PAGE=self.nb_videos_per_page
                )
            )

        # write list of videos in data.js
        def to_data_js(video):
            return {
                "id": video["contentDetails"]["videoId"],
                "title": video["snippet"]["title"],
                "slug": get_slug(video["snippet"]["title"]),
                "description": video["snippet"]["description"],
                "subtitles": get_subtitles(video["contentDetails"]["videoId"]),
                "thumbnail": str(
                    Path("videos").joinpath(
                        video["contentDetails"]["videoId"], "video.jpg"
                    )
                ),
            }

        with open(self.assets_dir.joinpath("data.js"), "w", encoding="utf-8") as fp:
            # write all playlists as they are
            for playlist in self.playlists:
                # retrieve list of videos for PL
                playlist_videos = load_json(
                    self.cache_dir, f"playlist_{playlist.playlist_id}_videos"
                )
                # filtering-out missing ones (deleted or not downloaded)
                playlist_videos = list(filter(skip_deleted_videos, playlist_videos))
                playlist_videos = list(filter(is_present, playlist_videos))
                # sorting them based on playlist
                playlist_videos.sort(key=lambda v: v["snippet"]["position"])

                fp.write(
                    "var json_{slug} = {json_str};\n".format(
                        slug=playlist.slug,
                        json_str=json.dumps(
                            list(map(to_data_js, playlist_videos)), indent=4
                        ),
                    )
                )

        # write a metadata.json file with some content-related data
        with open(
            self.build_dir.joinpath("metadata.json"), "w", encoding="utf-8"
        ) as fp:
            json.dump({"video_format": self.video_format}, fp, indent=4)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import pathlib
import subprocess

from .constants import logger
from .utils import nicer_args_join, resize_image


def hook_youtube_dl_ffmpeg(video_format, low_quality, data):
    """ youtube-dl hook to convert video at end of download

        - if low_quality was request
        - if received format is not requested one """

    if data.get("status") != "finished":
        return

    src_path = pathlib.Path(data["filename"])
    tmp_path = src_path.parent.joinpath(
        "video.tmp.{fmt}".format(src=src_path.name, fmt=video_format)
    )
    dst_path = src_path.parent.joinpath("video.{fmt}".format(fmt=video_format))

    # resize thumbnail. we use max width:248x187px in listing
    # but our posters are 480x270px
    resize_image(
        src_path.parent.joinpath("video.jpg"), width=480, height=270, method="cover"
    )

    # don't reencode if not requesting low-quality and received wanted format
    if not low_quality and src_path.suffix[1:] == video_format:
        return

    video_codecs = {"mp4": "h264", "webm": "libvpx"}
    audio_codecs = {"mp4": "aac", "webm": "libvorbis"}
    params = {"mp4": ["-movflags", "+faststart"], "webm": []}

    args = ["ffmpeg", "-y", "-i", "file:{}".format(str(src_path))]

    if low_quality:
        args += [
            "-codec:v",
            video_codecs[video_format],
            "-quality",
            "best",
            "-cpu-used",
            "0",
            "-b:v",
            "300k",
            "-qmin",
            "30",
            "-qmax",
            "42",
            "-maxrate",
            "300k",
            "-bufsize",
            "1000k",
            "-threads",
            "8",
            "-vf",
            "scale='480:trunc(ow/a/2)*2'",
            "-codec:a",
            audio_codecs[video_format],
            "-b:a",
            "128k",
        ]
    else:
        args += [
            "-codec:v",
            video_codecs[video_format],
            "-quality",
            "best",
            "-cpu-used",
            "0",
            "-bufsize",
            "1000k",
            "-threads",
            "8",
            "-codec:a",
            audio_codecs[video_format],
        ]
    args += params[video_format]
    args += ["file:{}".format(str(tmp_path))]

    logger.info(
        "  converting {src} into {dst}".format(src=str(src_path), dst=str(dst_path))
    )
    logger.debug(nicer_args_join(args))

    ffmpeg = subprocess.run(args)
    ffmpeg.check_returncode()

    # delete original
    src_path.unlink()
    # rename temp filename with final one
    tmp_path.replace(dst_path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import subprocess

from .constants import SCRAPER, logger
from .utils import nicer_args_join


class ZimInfo(object):
    def __init__(
        self,
        language="eng",
        title="my title",
        description="my zim description",
        creator="youtube",
        publisher="kiwix",
        name="youtube-test",
        tags=[],
        homepage="home.html",
        favicon="favicon.jpg",
        scraper=SCRAPER,
    ):

        self.homepage = homepage
        self.favicon = favicon
        self.language = language
        self.title = title
        self.description = description
        self.creator = creator
        self.publisher = publisher
        self.name = name
        self.tags = tags
        self.scraper = scraper

    def update(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def to_zimwriterfs_args(self):
        return [
            "--welcome",
            self.homepage,
            "--favicon",
            self.favicon,
            "--language",
            self.language,
            "--title",
            self.title,
            "--description",
            self.description,
            "--creator",
            self.creator,
            "--publisher",
            self.publisher,
            "--name",
            self.name,
            "--tags",
            ";".join(self.tags),
            "--scraper",
            self.scraper,
        ]


def make_zim_file(build_dir, output_dir, zim_fname, zim_info):
    """ runs zimwriterfs """
    args = (
        ["zimwriterfs"]
        + zim_info.to_zimwriterfs_args()
        + ["--verbose", str(build_dir), str(output_dir.joinpath(zim_fname))]
    )

    logger.debug(nicer_args_join(args))
    zimwriterfs = subprocess.run(args)
    zimwriterfs.check_returncode()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import pathlib
import subprocess
from setuptools import setup

from youtube2zim.constants import NAME, VERSION

ROOT_DIR = pathlib.Path(__file__).parent

with open(ROOT_DIR.joinpath("requirements.txt"), "r") as fp:
    requirements = [
        line.strip() for line in fp.readlines() if not line.strip().startswith("#")
    ]

with open(ROOT_DIR.joinpath("README.md"), "r") as fp:
    long_description = fp.read()

print("Downloading and fixing JS dependencies...")
ps = subprocess.run(["/bin/sh", str(ROOT_DIR.joinpath("get_js_deps.sh"))])
ps.check_returncode()


setup(
    name=NAME,
    version=VERSION,
    description="Make ZIM file from a Youtube channel, user or playlist(s)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="dattaz",
    author_email="taz@dattaz.fr",
    url="https://github.com/openzim/youtube",
    keywords="kiwix zim youtube offline",
    license="GPLv3+",
    packages=["youtube2zim"],
    install_requires=requirements,
    zip_safe=False,
    platforms="Linux",
    include_package_data=True,
    entry_points={"console_scripts": ["youtube2zim=youtube2zim.__main__:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
    python_requires=">=3.6",
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import logging
import argparse

from .constants import NAME, SCRAPER, CHANNEL, PLAYLIST, USER, logger, YOUTUBE
from .scraper import Youtube2Zim


def main():
    parser = argparse.ArgumentParser(
        prog=NAME,
        description="Scraper to create a ZIM file from a Youtube Channel or Playlists",
    )

    parser.add_argument(
        "--type",
        help="Type of collection",
        choices=[CHANNEL, PLAYLIST, USER],
        required=True,
        dest="collection_type",
    )
    parser.add_argument(
        "--id", help="Youtube ID of the collection", required=True, dest="youtube_id"
    )
    parser.add_argument("--api-key", help="Youtube API Token", required=True)
    parser.add_argument(
        "--name",
        help="ZIM name. Used as identifier and filename (date will be appended)",
        required=True,
    )

    parser.add_argument(
        "--format",
        help="Format to download/transcode video to. webm is smaller",
        choices=["mp4", "webm"],
        default="webm",
        dest="video_format",
    )
    parser.add_argument(
        "--low-quality",
        help="Re-encode video using stronger compression",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--all-subtitles",
        help="Include auto-generated subtitles",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--pagination",
        help="Number of videos per page",
        type=int,
        dest="nb_videos_per_page",
        default=40,
    )
    parser.add_argument(
        "--autoplay",
        help="Enable autoplay on video articles (home never have autoplay). Behavior differs on platforms/browsers.",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--output",
        help="Output folder for ZIM file or build folder",
        default="/output",
        dest="output_dir",
    )
    parser.add_argument(
        "--no-zim",
        help="Don't produce a ZIM file, create HTML folder only.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--zim-file",
        help="ZIM file name (based on --name if not provided)",
        dest="fname",
    )

    parser.add_argument(
        "--language", help="ISO-639-3 (3 chars) language code of content", default="eng"
    )
    parser.add_argument(
        "--locale",
        help="Locale name to use for translations (if avail) and time representations. Defaults to --language or English.",
        dest="locale_name",
    )
    parser.add_argument(
        "--title",
        help="Custom title for your project and ZIM. Default to Channel name (of first video if playlists)",
    )
    parser.add_argument(
        "--description",
        help="Custom description for your project and ZIM. Default to Channel name (of first video if playlists)",
    )
    parser.add_argument(
        "--creator",
        help="Name of content creator. Defaults to Channel name or “Youtue Channels”",
    )
    parser.add_argument(
        "--publisher", help="Custom publisher name (ZIM metadata)", default="Kiwix"
    )
    parser.add_argument(
        "--tags",
        help="List of comma-separated Tags for the ZIM file. _videos:yes added automatically",
        default="youtube",
    )
    parser.add_argument(
        "--profile",
        help="Custom profile image. Squared. Will be resized to 100x100px",
        dest="profile_image",
    )
    parser.add_argument(
        "--banner",
        help="Custom banner image. Will be resized to 1060x175px",
        dest="banner_image",
    )
    parser.add_argument(
        "--main-color",
        help="Custom color. Hex/HTML syntax (#DEDEDE). Default to main color of profile image.",
    )
    parser.add_argument(
        "--secondary-color",
        help="Custom secondary color. Hex/HTML syntax (#DEDEDE). Default to secondary color of profile image.",
    )

    parser.add_argument(
        "--debug", help="Enable verbose output", action="store_true", default=False
    )
    parser.add_argument(
        "--keep",
        help="Don't erase build folder on start (for debug/devel)",
        default=False,
        action="store_true",
        dest="keep_build_dir",
    )
    parser.add_argument(
        "--skip-download",
        help="Skip the download step (videos, thumbnails, subtitles)",
        default=False,
        action="store_true",
        dest="skip_download",
    )
    parser.add_argument(
        "--concurrency",
        help="Number of concurrent threads to use",
        type=int,
        dest="max_concurrency",
        default=1,
    )
    parser.add_argument(
        "--only_test_branding",
        help="Just generate a fake home.html to check branding (images and colors)",
        default=False,
        action="store_true",
        dest="only_test_branding",
    )
    parser.add_argument(
        "--version",
        help="Display scraper version and exit",
        action="version",
        version=SCRAPER,
    )

    args = parser.parse_args()
    logger.setLevel(logging.DEBUG if args.debug else logging.INFO)

    try:
        if args.max_concurrency < 1:
            raise ValueError(f"Invalid concurrency value: {args.max_concurrency}")
        scraper = Youtube2Zim(**dict(args._get_kwargs()), youtube_store=YOUTUBE)
        scraper.run()
    except Exception as exc:
        logger.error(f"FAILED. An error occured: {exc}")
        if args.debug:
            logger.exception(exc)
        raise SystemExit(1)


if __name__ == "__main__":
    main()
